/* Simple object shadows */
/* An OpenGL demo by A.P.Gardner */

#include <GL/glut.h>
#include <stdio.h>
#include <stdlib.h>

/* definitions */

#define YELLOWMAT 1
#define REDMAT 2
#define BLUEMAT 3

/* variables */

GLuint texName;		/* store of shadow texture reference */

/* Function to load a texture from a bmp file */

void makeTextureImage(char filename[], int size)
{
	long data_start = 54;			/* start of data section in files wall.bmp & light.bmp */
									/* NB header size of bmp files is not fixed, refer to M$ documentation */
	unsigned char ch, cho[4];		/* temp file input buffers */
	FILE *fileptr;					/* pointer to bmp file */
	GLubyte image128[128][128][3];	/* storage for 128x128 texture */
	int n, m, i, j;					/* counters */

	/* open file */

	if ((fileptr = fopen(filename, "rb")) == NULL)
		return;

	/* step over the header */

	n = 0;
	while (!feof(fileptr) && n < data_start)
	{
		ch = fgetc(fileptr);
		n++;
	}

	/* load the texture data */

	for (i=0; i<size; i++)
	{
		for (j=0; j<size; j++)
		{
			m=0;
			while(!feof(fileptr) && m < 3)
			{
				cho[m] = fgetc(fileptr);
				m++;
			}
			image128[i][j][0] = cho[2];	/* bmp using BGR values, hence the swap */
			image128[i][j][1] = cho[1];
			image128[i][j][2] = cho[0];
		}
	}

	/* close the file */

	fclose(fileptr);

	/* create texture object */
	
	glBindTexture(GL_TEXTURE_2D, texName);

	/* set up wrap & filter params */

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	/* load the image */

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, size, size,
				 0, GL_RGB, GL_UNSIGNED_BYTE, image128);
}

/* Function to initialise params */

void init(void)
{
	GLfloat yellow_diffuse[] = {0.7, 0.7, 0.0, 1.0};
	GLfloat yellow_specular[] = {1.0, 1.0, 1.0, 1.0};
	GLfloat red_diffuse[] = {1.0, 0.0, 0.0, 1.0};
	GLfloat red_specular[] = {1.0, 1.0, 1.0, 1.0};
	GLfloat blue_diffuse[] = {0.1, 0.1, 0.7, 1.0};
	GLfloat blue_specular[] = {1.0, 1.0, 1.0, 1.0};
	GLfloat ambient_light[] = {0.5, 0.5, 0.5, 1.0};
	GLfloat position_one[] = {0.0, 0.9, 0.4, 0.0};

	/* create materials for objects in the scene */

	glNewList(YELLOWMAT, GL_COMPILE);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, yellow_diffuse);
	glMaterialfv(GL_FRONT, GL_SPECULAR, yellow_specular);
	glMaterialf(GL_FRONT, GL_SHININESS, 64.0);
	glColor3fv(yellow_diffuse);
	glEndList();

	glNewList(REDMAT, GL_COMPILE);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, red_diffuse);
	glMaterialfv(GL_FRONT, GL_SPECULAR, red_specular);
	glMaterialf(GL_FRONT, GL_SHININESS, 64.0);
	glColor3fv(red_diffuse);
	glEndList();

	glNewList(BLUEMAT, GL_COMPILE);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, blue_diffuse);
	glMaterialfv(GL_FRONT, GL_SPECULAR, blue_specular);
	glMaterialf(GL_FRONT, GL_SHININESS, 45.0);
	glColor3fv(blue_diffuse);
	glEndList();

	/* generate texture names */

	texName = 0;
	glGenTextures(1, &texName);

	/* load textures */

	makeTextureImage("shadow.bmp", 128);

	/* set lighting params */

	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient_light);
	glLightfv(GL_LIGHT0, GL_POSITION, position_one);
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHTING);

	/* enable depth buffering */

	glEnable(GL_DEPTH_TEST);

	/* set texturing function */

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}

/* Function to clean up */

void cleanup(void)
{
	/* delete texture objects */

	glDeleteTextures(1, &texName);
}

/* Function to draw object and its shadow */

void draw_object()
{
	/* draw object */

	glCallList(BLUEMAT);
	glutSolidCone(0.5, 1.0, 30, 30);

	/* set new render state */

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_ZERO, GL_SRC_COLOR);
	glDepthMask(GL_FALSE);
	glDepthFunc(GL_LEQUAL);

	/* translate to shadow position & render shadow panel */

	glTranslatef(0.0, -0.95, 0.0);
	glBindTexture(GL_TEXTURE_2D, texName);
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0);
		glVertex3f(-0.7, 0.0, 0.7);
		glTexCoord2f(1.0, 0.0);
		glVertex3f(0.7, 0.0, 0.7);
		glTexCoord2f(1.0, 1.0);
		glVertex3f(0.7, 0.0, -0.7);
		glTexCoord2f(0.0, 1.0);
		glVertex3f(-0.7, 0.0, -0.7);
	glEnd();

	/* reset rendering state to default */

	glDepthFunc(GL_LESS);
	glDepthMask(GL_TRUE);
	glDisable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);
}

/* Function to draw display */

void display(void)
{
	/* clear all buffers */

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/* draw floor & wall panel */

	glStencilFunc(GL_ALWAYS, 0, 0);
	glCallList(YELLOWMAT);
	glBegin(GL_QUADS);
		glNormal3f(0.0, 1.0, 0.0);
		glVertex3f(3.0, -1.0, 4.5);
		glVertex3f(3.0, -1.0, -5.0);
		glVertex3f(-3.0, -1.0, -5.0);
		glVertex3f(-3.0, -1.0, 4.5);
	glEnd();
	glCallList(REDMAT);
	glBegin(GL_QUADS);
		glNormal3f(0.0, 0.0, 1.0);
		glVertex3f(-3.0, 1.0, 1.0);
		glVertex3f(-1.0, 1.0, 1.0);
		glVertex3f(-1.0, -1.0, 1.0);
		glVertex3f(-3.0, -1.0, 1.0);
	glEnd();

	/* draw 1st object & shadow */

	glPushMatrix();
	glTranslatef(-1.0, 0.0, 0.0);
	draw_object();
	glPopMatrix();

	/* draw 2nd object & shadow */

	glPushMatrix();
	glTranslatef(1.0, 0.0, 0.0);
	draw_object();
	glPopMatrix();

	/* process all buffered OpenGL commands */

	glFlush();
}

/* Function to handle window size changes */

void reshape(int w, int h)
{
	glViewport(0, 0, (GLsizei)w, (GLsizei)h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0, (GLfloat)w/(GLfloat)h, 3.0, 7.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0.0, 0.0, -5.0);
}

/* Main function */

int main(int argc, char** argv)
{
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH | GLUT_SINGLE | GLUT_RGB);
	glutInitWindowSize(500, 500);
	glutInitWindowPosition(100, 100);
	glutCreateWindow("OpenGL Demo - Simple Shadows");
	init();
	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutMainLoop();
	cleanup();

	return 0;
}